/*
 * File access convenience functions.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <string.h>

#include "dataspace_client.h"
#include "file_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "pipe_client.h"
#include "pipe_opener_client.h"
#include "mapped_file_client.h"

#include "file.h"



/* Update the extent of the file in a region using the region start and end
   positions and the file size. */

static void _update_extent(file_t *file)
{
  /* Handle files ending after or within the region. */

  if (file->size > file->start_pos)
  {
    if (file->size > file->end_pos)
      file->data_end = file->end_pos - file->start_pos;
    else
      file->data_end = file->size - file->start_pos;
  }

  /* Handle files ending before the region. */

  else
    file->data_end = 0;
}



/* Initialise the given file structure. */

void file_init(file_t *file)
{
  file->memory = NULL;
  file->ref = L4_INVALID_CAP;
  file->start_pos = 0;
  file->end_pos = 0;
  file->data_end = 0;
  file->data_current = 0;
  file->can_mmap = 1;
  file->has_size = 1;
}



/* Release resources for the given file. */

void file_close(file_t *file)
{
  if (l4_is_valid_cap(file->ref))
    ipc_cap_free_um(file->ref);

  if (file->memory != NULL)
    ipc_detach_dataspace(file->memory);

  file_init(file);
}

/* Open a file using the given structure, indicating the filename and
   filesystem server. The file_mmap function should be used to obtain access to
   memory providing file data. This is a convenience function invoking
   file_context and file_context_open. */

long file_open(file_t *file, const char *filename, flags_t flags, l4_cap_idx_t server)
{
  file_t context;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, NULL))
    return -L4_ENOMEM;

  err = file_context_open(file, flags, &context);

  /* Close the context, although a separate mechanism could permit contexts to
     open several files. */

  file_close(&context);
  return err;
}



/* Initialise a file structure for a context obtained from the given server
   attaching memory to communicate filename information. */

long file_context(file_t *file, l4_cap_idx_t server)
{
  if (l4_is_invalid_cap(server))
    return -L4_EINVAL;

  client_Opener opener(server);
  offset_t size;
  flags_t flags;
  long err;

  file_init(file);

  err = opener.context(&file->ref);
  if (err)
    return err;

  client_Dataspace context_ds(file->ref);

  err = context_ds.info(&size, &flags);
  if (err)
    return err;

  file->start_pos = 0;
  file->end_pos = size;

  return ipc_attach_dataspace(file->ref, size, (void **) &file->memory);
}

/* Open a file using the given structure and context. */

long file_context_open(file_t *file, flags_t flags, file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  file_init(file);
  return openercontext.open(flags, &file->size, &file->ref);
}



/* Flush populated data and obtain an updated file size and populated data
   details. */

long file_flush(file_t *file)
{
  client_File _file(file->ref);
  long err = _file.flush(file->data_current, &file->size);

  if (err)
    return err;

  _update_extent(file);

  return L4_EOK;
}

/* Map a region of the given file to a memory region, obtaining an updated file
   size and populated data details. Unmap any previously mapped region. */

long file_mmap(file_t *file, offset_t position, offset_t length)
{
  char *memory = file->memory;
  client_MappedFile mapped_file(file->ref);
  long err = mapped_file.mmap(position, length, &file->start_pos,
                              &file->end_pos, &file->size);

  if (err)
    return err;

  _update_extent(file);

  err = ipc_attach_dataspace(file->ref, file_span(file), (void **) &file->memory);
  if (err)
    return err;

  if (memory != NULL)
    ipc_detach_dataspace(memory);

  return L4_EOK;
}

/* Resize a file, obtaining updated file size and populated data details. */

long file_resize(file_t *file, offset_t size)
{
  if (!file->has_size)
    return -L4_EIO;

  client_File _file(file->ref);
  offset_t file_size = size;
  long err = _file.resize(&file_size);

  if (err)
    return err;

  file->size = file_size;
  _update_extent(file);
  return L4_EOK;
}



/* Return the amount of data in the mapped region for the given file. */

offset_t file_populated_span(file_t *file)
{
  offset_t size = file_span(file);
  return (file->data_end < size) ? file->data_end : size;
}

/* Return the size of the mapped region for the given file. */

offset_t file_span(file_t *file)
{
  return file->end_pos - file->start_pos;
}



/* Get a pointer to any terminated string at the given offset or NULL if the
   data from offset is not terminated. */

char *file_string_get(file_t *file, offset_t offset)
{
  offset_t limit = file_span(file) - offset;

  if (strnlen(file->memory + offset, limit) < limit)
    return file->memory + offset;
  else
    return NULL;
}

/* Copy a string to the mapped region at the given offset, returning 1 (true)
   where all characters were copied, 0 (false) otherwise. The precise number of
   characters copied, excluding the zero terminator is provided via the written
   parameter if it is not specified as NULL. */

int file_string_set(file_t *file, const char *data, offset_t offset,
                    offset_t *written)
{
  offset_t i, pos, limit = file_span(file);

  /* Do not attempt to copy data with an invalid offset. */

  if (offset >= limit)
  {
    if (written != NULL)
      *written = 0;
    return 0;
  }

  /* Copy the data to the given offset, stopping at the end of the region. */

  for (i = 0, pos = offset; pos < limit; i++, pos++)
  {
    file->memory[pos] = data[i];

    /* Terminator written, can return immediately. */

    if (!data[i])
    {
      if (written != NULL)
        *written = pos - offset;
      return 1;
    }
  }

  /* Terminate the incomplete string at the end of the region. */

  file->memory[limit - 1] = '\0';
  if (written != NULL)
    *written = limit - 1 - offset;
  return 0;
}



/* Return the number of remaining populated bytes in the region. */

offset_t file_data_available(file_t *file)
{
  return file_populated_span(file) - file->data_current;
}

/* Return the current data offset in the region. */

char *file_data_current(file_t *file)
{
  return file->memory + file->data_current;
}

/* Return the current access position in the file. */

offset_t file_data_current_position(file_t *file)
{
  return file->start_pos + file->data_current;
}

/* Return the position of the end of the populated bytes in the region. */

offset_t file_data_end_position(file_t *file)
{
  return file->start_pos + file->data_end;
}

/* Return the amount of remaining space in the region. */

offset_t file_data_space(file_t *file)
{
  return file_span(file) - file->data_current;
}



/* Copy data to the given buffer from the current data position, updating the
   position. */

void file_data_read(file_t *file, char *buf, offset_t to_transfer)
{
  memcpy(buf, file_data_current(file), to_transfer);

  /* Update position details. */

  file->data_current += to_transfer;
}

/* Copy data from the given buffer to the current data position, updating the
   position and the extent of populated data if this was exceeded. */

void file_data_write(file_t *file, char *buf, offset_t to_transfer)
{
  memcpy(file_data_current(file), buf, to_transfer);

  /* Update position details. */

  file->data_current += to_transfer;

  if (file->data_current > file->data_end)
    file->data_end = file->data_current;
}



/* Open two pipe endpoints using the given pipe server. */

long pipe_open(offset_t size, file_t *reader, file_t *writer, l4_cap_idx_t server)
{
  if (l4_is_invalid_cap(server))
    return -L4_EINVAL;

  client_PipeOpener opener(server);

  file_init(reader);
  file_init(writer);

  /* Pipes can usually only be accessed via region navigation. */

  reader->can_mmap = 0;
  reader->has_size = 0;
  writer->can_mmap = 0;
  writer->has_size = 0;

  long err = opener.pipe(size, &reader->ref, &writer->ref);
  if (err)
    return err;

  err = pipe_next(writer) || pipe_next(reader);

  if (err)
  {
    file_close(reader);
    file_close(writer);
  }

  return err;
}

/* Access the current region for a pipe endpoint. */

long pipe_current(file_t *pipe)
{
  client_Pipe _pipe(pipe->ref);
  long err = _pipe.current_region(&pipe->data_end, &pipe->size);
  char *memory = pipe->memory;

  if (err)
    return err;

  pipe->end_pos = pipe->size;

  err = ipc_attach_dataspace(pipe->ref, file_span(pipe), (void **) &pipe->memory);
  if (err)
    return err;

  if (memory != NULL)
    ipc_detach_dataspace(memory);

  return L4_EOK;
}

/* Access the next region for a pipe endpoint, updating the eventual size of
   the current region. */

long pipe_next(file_t *pipe)
{
  client_Pipe _pipe(pipe->ref);
  long err = _pipe.next_region(&pipe->data_end, &pipe->size);
  char *memory = pipe->memory;

  if (err)
    return err;

  pipe->end_pos = pipe->size;

  err = ipc_attach_dataspace(pipe->ref, file_span(pipe), (void **) &pipe->memory);
  if (err)
    return err;

  if (memory != NULL)
    ipc_detach_dataspace(memory);

  return L4_EOK;
}

/* Set the size of the written region. */

long pipe_written(file_t *pipe, offset_t size)
{
  if (size <= pipe->size)
  {
    pipe->data_end = size;
    return L4_EOK;
  }
  else
    return -L4_EINVAL;
}

// vim: tabstop=2 expandtab shiftwidth=2
